# -*- coding: utf-8 -*-

"""Tests of mock datasets."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_array_equal as ae

from ..mock import (artificial_waveforms,
                    artificial_traces,
                    artificial_spike_clusters,
                    artificial_features,
                    artificial_masks,
                    artificial_spike_samples,
                    artificial_correlograms,
                    )


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def _test_artificial(n_spikes=None, n_clusters=None):
    n_samples_waveforms = 32
    n_samples_traces = 50
    n_channels = 35
    n_features = n_channels * 2

    # Waveforms.
    waveforms = artificial_waveforms(n_spikes=n_spikes,
                                     n_samples=n_samples_waveforms,
                                     n_channels=n_channels)
    assert waveforms.shape == (n_spikes, n_samples_waveforms, n_channels)

    # Traces.
    traces = artificial_traces(n_samples=n_samples_traces,
                               n_channels=n_channels)
    assert traces.shape == (n_samples_traces, n_channels)

    # Spike clusters.
    spike_clusters = artificial_spike_clusters(n_spikes=n_spikes,
                                               n_clusters=n_clusters)
    assert spike_clusters.shape == (n_spikes,)
    if n_clusters >= 1:
        assert spike_clusters.min() in (0, 1)
        assert spike_clusters.max() in (n_clusters - 1, n_clusters - 2)
    ae(np.unique(spike_clusters), np.arange(n_clusters))

    # Features.
    features = artificial_features(n_spikes, n_features)
    assert features.shape == (n_spikes, n_features)

    # Masks.
    masks = artificial_masks(n_spikes, n_channels)
    assert masks.shape == (n_spikes, n_channels)

    # Spikes.
    spikes = artificial_spike_samples(n_spikes)
    assert spikes.shape == (n_spikes,)

    # CCG.
    ccg = artificial_correlograms(n_clusters, 10)
    assert ccg.shape == (n_clusters, n_clusters, 10)


def test_artificial():
    _test_artificial(n_spikes=100, n_clusters=10)
    _test_artificial(n_spikes=0, n_clusters=0)



# -*- coding: utf-8 -*-

"""Tests of cluster statistics."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_array_equal as ae
from numpy.testing import assert_allclose as ac
from pytest import fixture

from ..clusters import (mean,
                        get_unmasked_channels,
                        get_mean_probe_position,
                        get_sorted_main_channels,
                        get_mean_masked_features_distance,
                        get_waveform_amplitude,
                        )
from phylib.utils.geometry import staggered_positions
from phylib.io.mock import artificial_features, artificial_masks, artificial_waveforms


#------------------------------------------------------------------------------
# Fixtures
#------------------------------------------------------------------------------

@fixture
def n_channels():
    yield 28


@fixture
def n_spikes():
    yield 50


@fixture
def n_samples():
    yield 40


@fixture
def n_features_per_channel():
    yield 4


@fixture
def features(n_spikes, n_channels, n_features_per_channel):
    yield artificial_features(n_spikes, n_channels, n_features_per_channel)


@fixture
def masks(n_spikes, n_channels):
    yield artificial_masks(n_spikes, n_channels)


@fixture
def waveforms(n_spikes, n_samples, n_channels):
    yield artificial_waveforms(n_spikes, n_samples, n_channels)


@fixture
def site_positions(n_channels):
    yield staggered_positions(n_channels)


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def test_mean(features, n_channels, n_features_per_channel):
    mf = mean(features)
    assert mf.shape == (n_channels, n_features_per_channel)
    ae(mf, features.mean(axis=0))


def test_unmasked_channels(masks, n_channels):
    # Mask many values in the masks array.
    threshold = .05
    masks[:, 1::2] *= threshold
    # Compute the mean masks.
    mean_masks = mean(masks)
    # Find the unmasked channels.
    channels = get_unmasked_channels(mean_masks, threshold)
    # These are 0, 2, 4, etc.
    ae(channels, np.arange(0, n_channels, 2))


def test_mean_probe_position(masks, site_positions):
    masks[:, ::2] *= .05
    mean_masks = mean(masks)
    mean_pos = get_mean_probe_position(mean_masks, site_positions)
    assert mean_pos.shape == (2,)
    assert mean_pos[0] < 0
    assert mean_pos[1] > 0


def test_sorted_main_channels(masks):
    masks *= .05
    masks[:, [5, 7]] *= 20
    mean_masks = mean(masks)
    channels = get_sorted_main_channels(mean_masks,
                                        get_unmasked_channels(mean_masks))
    assert np.all(np.isin(channels, [5, 7]))


def test_waveform_amplitude(masks, waveforms):
    waveforms *= .1
    masks *= .1

    waveforms[:, 10, :] *= 10
    masks[:, 10] *= 10

    mean_waveforms = mean(waveforms)
    mean_masks = mean(masks)

    amplitude = get_waveform_amplitude(mean_masks, mean_waveforms)
    assert np.all(amplitude >= 0)
    assert amplitude.shape == (mean_waveforms.shape[1],)


def test_mean_masked_features_distance(features,
                                       n_channels,
                                       n_features_per_channel,
                                       ):

    # Shifted feature vectors.
    shift = 10.
    f0 = mean(features)
    f1 = mean(features) + shift

    # Only one channel is unmasked.
    m0 = m1 = np.zeros(n_channels)
    m0[n_channels // 2] = 1

    # Check the distance.
    d_expected = np.sqrt(n_features_per_channel) * shift
    d_computed = get_mean_masked_features_distance(f0, f1, m0, m1,
                                                   n_features_per_channel)
    ac(d_expected, d_computed)



# -*- coding: utf-8 -*-

"""Probe merging."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging
from pathlib import Path

from tqdm import tqdm
import numpy as np
from scipy.linalg import block_diag

from phylib.utils._misc import (
    _read_tsv_simple, _write_tsv_simple, write_tsv, read_python, write_python)
from phylib.io.model import load_model

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Merge utils
#------------------------------------------------------------------------------

def _concat(arrs, axis=0, dtype=None):
    dtype = dtype or arrs[0].dtype
    return np.concatenate(arrs).astype(dtype)


def _load_multiple_spike_times(*spike_times_l):
    """Load multiple spike_times arrays and merge them into a single one."""
    # We concatenate all spike times arrays.
    spike_times_concat = _concat(spike_times_l)
    # We sort by increasing time.
    spike_order = np.argsort(spike_times_concat, kind='stable')
    spike_times_ordered = spike_times_concat[spike_order]
    assert np.all(np.diff(spike_times_ordered) >= 0)
    # We return the ordered spike times, and the reordering array.
    return spike_times_ordered, spike_order


def _load_multiple_spike_arrays(*spike_array_l, spike_order=None):
    """Load multiple spike-dependent arrays and concatenate them along the first dimension.
    Keep the spike time ordering.
    """
    assert spike_order is not None
    spike_array_concat = _concat(spike_array_l, axis=0)
    assert spike_array_concat.shape[0] == spike_order.shape[0]
    return spike_array_concat[spike_order]


def _load_multiple_files(fn, subdirs):
    """Load the same filename in the different subdirectories."""
    # Warning: squeeze may fail in degenerate cases.
    return [np.load(str(subdir / fn)).squeeze() for subdir in subdirs]


#------------------------------------------------------------------------------
# Main Merger class
#------------------------------------------------------------------------------

class Merger(object):
    """Merge spike-sorted data from different probes and output datasets to disk.

    Constructor
    -----------

    subdirs : list
        List of paths to the probe directories.
    out_dir : str or Path
        Output directory for the merged spike-sorted data.
    probe_info : list
        For each probe, a dictionary with the following fields:
            * `label`: a string with the probe label (defaults to the probe folder name)
            * any other field, such as `model`, `serial_number`, etc.

        All fields will be saved in `probes.description.tsv`.

    """
    def __init__(self, subdirs, out_dir, probe_info=None):
        assert subdirs
        self.subdirs = [Path(subdir) for subdir in subdirs]

        self.out_dir = Path(out_dir)
        self.out_dir.mkdir(parents=True, exist_ok=True)

        # Default probe info if not provided: the label is the probe folder name.
        self.probe_info = probe_info or [{'label': subdir.parts[-1]} for subdir in self.subdirs]
        assert len(self.probe_info) == len(self.subdirs)

    def _save(self, name, arr):
        """Save a npy array in the output directory."""
        logger.debug("Saving %s %s %s.", name, arr.dtype, arr.shape)
        np.save(self.out_dir / name, arr)

    def write_params(self):
        """Write a params.py for the merged dataset."""
        params_l = [read_python(subdir / 'params.py') for subdir in self.subdirs]
        n_channels_dat = sum(params['n_channels_dat'] for params in params_l)

        params_merged = params_l[0]
        params_merged['dat_path'] = []
        params_merged['n_channels_dat'] = n_channels_dat

        write_python(self.out_dir / 'params.py', params_merged)

    def write_probe_desc(self):
        """Write the probe description in a TSV file."""
        write_tsv(self.out_dir / 'probes.description.tsv', self.probe_info, first_field='label')

    def write_spike_times(self):
        """Write the merged spike times, and register self.spike_order with the reordering
        of the spikes."""
        spike_times_l = _load_multiple_files('spike_times.npy', self.subdirs)
        spike_times, self.spike_order = _load_multiple_spike_times(*spike_times_l)
        self._save('spike_times.npy', spike_times)

    def write_spike_data(self):
        """Write spike-dependent data."""
        spike_data = [
            'amplitudes.npy',
            'spike_templates.npy',
            # 'pc_features.npy',
            # 'template_features.npy',
        ]
        for fn in spike_data:
            arrays = _load_multiple_files(fn, self.subdirs)
            concat = _load_multiple_spike_arrays(*arrays, spike_order=self.spike_order)
            self._save(fn, concat)

    def write_spike_clusters(self):
        """Write the merged spike clusters, and register self.cluster_offsets.
           Write the merged spike templates, and register self.template_offsets.
        """
        spike_clusters_l = _load_multiple_files('spike_clusters.npy', self.subdirs)
        spike_templates_l = _load_multiple_files('spike_templates.npy', self.subdirs)
        self.cluster_offsets = []
        self.template_offsets = []
        cluster_probes_l = []
        coffset = 0
        toffset = 0
        for i, (subdir, sc, st) in enumerate(
                zip(self.subdirs, spike_clusters_l, spike_templates_l)):
            n_clu = np.max(sc) + 1
            n_tmp = np.max(st) + 1
            sc += coffset
            st += toffset
            self.cluster_offsets.append(coffset)
            self.template_offsets.append(toffset)
            cluster_probes_l.append(i * np.ones(n_clu, dtype=np.int32))
            coffset += n_clu
            toffset += n_tmp
        spike_clusters = _load_multiple_spike_arrays(
            *spike_clusters_l, spike_order=self.spike_order)
        spike_templates = _load_multiple_spike_arrays(
            *spike_templates_l, spike_order=self.spike_order)
        cluster_probes = _concat(cluster_probes_l)
        assert np.max(spike_clusters) + 1 == cluster_probes.size
        self._save('spike_clusters.npy', spike_clusters)
        self._save('spike_templates.npy', spike_templates)
        self._save('cluster_probes.npy', cluster_probes)

    def write_cluster_data(self):
        """We load all cluster metadata from TSV files, renumber the clusters,
        merge the dictionaries, and save in a new merged TSV file. """

        cluster_data = [
            'cluster_Amplitude.tsv',
            'cluster_ContamPct.tsv',
            'cluster_KSLabel.tsv'
        ]

        for fn in cluster_data:
            metadata = {}
            for subdir, offset in zip(self.subdirs, self.cluster_offsets):
                try:
                    field_name, metadata_loc = _read_tsv_simple(subdir / fn)
                except ValueError:
                    # Skipping non-existing file.
                    continue
                for k, v in metadata_loc.items():
                    metadata[k + offset] = v
            if metadata:
                _write_tsv_simple(self.out_dir / fn, field_name, metadata)

    def write_channel_data(self):
        """Write channel-dependent data, and register self.channel_offsets."""
        self.channel_offsets = []
        channel_probes = []
        channel_maps_l = _load_multiple_files('channel_map.npy', self.subdirs)
        # TODO if needed: channel_shanks.npy
        offset = 0
        for ind, array in enumerate(channel_maps_l):
            array += offset
            self.channel_offsets.append(offset)
            offset = array.max()
            channel_probes.append(array * 0 + ind)
        channel_maps = _concat(channel_maps_l, axis=0)
        channel_probes = _concat(channel_probes, axis=0)
        self._save('channel_map.npy', channel_maps)
        self._save('channel_probe.npy', channel_probes)

    def write_channel_positions(self):
        """Write the channel positions."""
        channel_positions_l = _load_multiple_files('channel_positions.npy', self.subdirs)
        x_offset = 0.
        for array in channel_positions_l:
            array[:, 0] += x_offset
            x_offset = 2. * array[:, 0].max() - array[:, 0].min()
        channel_positions = _concat(channel_positions_l, axis=0)
        self._save('channel_positions.npy', channel_positions)

    def write_templates(self):
        """Write the templates (only dense format for now)."""
        # TODO: write the templates array in sparse format.

        path = self.out_dir / 'templates.npy'

        templates_l = _load_multiple_files('templates.npy', self.subdirs)

        # Determine the templates array shape.
        n_templates = sum(tmp.shape[0] for tmp in templates_l)
        n_samples = templates_l[0].shape[1]  # assuming all have the same number of samples
        assert np.all(np.array([templates_i.shape[1] for templates_i in templates_l]) == n_samples)

        n_channels = sum(tmp.shape[2] for tmp in templates_l)
        shape = (n_templates, n_samples, n_channels)

        np.save(path, np.empty(shape, dtype=templates_l[0].dtype))
        offset = 0
        with open(path, 'r+b') as fid:
            fid.seek(8)
            offset = int.from_bytes(fid.read(2), byteorder='little')
            fid.seek(offset, 1)
            for i in range(len(self.subdirs)):
                j0 = templates_l[i - 1].shape[2] if i > 0 else 0
                j1 = j0 + templates_l[i].shape[2]
                for it in np.arange(templates_l[i].shape[0]):
                    one_template = np.zeros((n_samples, n_channels), dtype=templates_l[0].dtype)
                    one_template[:, j0:j1] = templates_l[i][it, :]
                    fid.write(one_template.tobytes())

    def write_template_data(self):
        template_data = [
            # 'templates_ind.npy',  # HACK: do not copy this array (which is trivial with 0 1 2 3..
            # on each row),
            # the templates.npy file is really dense in KS2 and should stay this way
            'pc_feature_ind.npy',
            'template_feature_ind.npy',
        ]

        for fn in template_data:
            arrays = _load_multiple_files(fn, self.subdirs)
            # For ind arrays, we need to take into account the channel offset.
            for array, offset in zip(arrays, self.channel_offsets):
                array += offset
            concat = _concat(arrays, axis=0).astype(np.uint32)
            self._save(fn, concat)

    def write_misc(self):
        """Write misc merged data.

        Similar templates: we make a block diagonal matrix from the n_templates * n_templates
        matrices, assuming no similarity between templates from different probes.

        Whitening matrix: same thing, except that the matrices are n_channels * n_channels.

        """
        diag_data = [
            'similar_templates.npy',
            'whitening_mat.npy',
            'whitening_mat_inv.npy',
        ]
        for fn in diag_data:
            try:
                concat = block_diag(*_load_multiple_files(fn, self.subdirs))
            except FileNotFoundError:
                logger.debug("File %s not found, skipping.", fn)
                continue
            self._save(fn, concat)

    def merge(self):
        """Merge the probes data and return a TemplateModel instance of the merged data."""

        with tqdm(desc="Merging", total=100) as bar:
            self.write_params()
            self.write_probe_desc()
            bar.update(10)
            self.write_spike_times()
            bar.update(10)
            self.write_spike_data()
            bar.update(10)
            self.write_spike_clusters()
            bar.update(10)
            self.write_cluster_data()
            bar.update(10)
            self.write_channel_data()
            bar.update(10)
            self.write_channel_positions()
            bar.update(10)
            self.write_templates()
            bar.update(10)
            self.write_template_data()
            bar.update(10)
            self.write_misc()
            bar.update(10)

        return load_model(self.out_dir / 'params.py')



# -*- coding: utf-8 -*-

"""Testing the Template model."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging

import numpy as np
from numpy.testing import assert_equal as ae
from pytest import raises

# from phylib.utils import Bunch
from phylib.utils.testing import captured_output
from ..model import from_sparse, load_model

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def test_from_sparse():
    data = np.array([[0, 1, 2], [3, 4, 5]])
    cols = np.array([[20, 23, 21], [21, 19, 22]])

    def _test(channel_ids, expected):
        expected = np.asarray(expected)
        dense = from_sparse(data, cols, np.array(channel_ids))
        assert dense.shape == expected.shape
        ae(dense, expected)

    _test([0], np.zeros((2, 1)))
    _test([19], [[0], [4]])
    _test([20], [[0], [0]])
    _test([21], [[2], [3]])

    _test([19, 21], [[0, 2], [4, 3]])
    _test([21, 19], [[2, 0], [3, 4]])

    with raises(NotImplementedError):
        _test([19, 19], [[0, 0], [4, 4]])


def test_model_1(template_model_full):
    with captured_output() as (stdout, stderr):
        template_model_full.describe()
    out = stdout.getvalue()
    assert 'sim_binary.dat' in out
    assert '64' in out


def test_model_2(template_model_full):
    m = template_model_full
    tmp = m.get_template(3)
    channel_ids = tmp.channel_ids
    spike_ids = m.get_cluster_spikes(3)

    w = m.get_waveforms(spike_ids, channel_ids)
    assert w is None or w.shape == (len(spike_ids), tmp.template.shape[0], len(channel_ids))

    f = m.get_features(spike_ids, channel_ids)
    assert f is None or f.shape == (len(spike_ids), len(channel_ids), 3)

    tf = m.get_template_features(spike_ids)
    assert tf is None or tf.shape == (len(spike_ids), m.n_templates)


def test_model_3(template_model_full):
    m = template_model_full

    spike_ids = m.get_template_spikes(3)
    n_spikes = len(spike_ids)

    channel_ids = m.get_template_channels(3)
    n_channels = len(channel_ids)

    waveforms = m.get_template_spike_waveforms(3)
    if waveforms is not None:
        assert waveforms.ndim == 3
        assert waveforms.shape[0] == n_spikes
        assert waveforms.shape[2] == n_channels

    tw = m.get_template_waveforms(3)
    assert tw.ndim == 2
    assert tw.shape[1] == n_channels


def test_model_4(template_model_full):
    m = template_model_full

    spike_ids = m.get_cluster_spikes(3)
    n_spikes = len(spike_ids)

    channel_ids = m.get_cluster_channels(3)
    n_channels = len(channel_ids)

    waveforms = m.get_cluster_spike_waveforms(3)
    if waveforms is not None:
        assert waveforms.ndim == 3
        assert waveforms.shape[0] == n_spikes
        assert waveforms.shape[2] == n_channels

    mean_waveforms = m.get_cluster_mean_waveforms(3)
    assert mean_waveforms.mean_waveforms.shape[1] == len(mean_waveforms.channel_ids)


def test_model_depth(template_model):
    depths = template_model.get_depths()
    assert depths.shape == (template_model.n_spikes,)


def test_model_merge(template_model_full):
    m = template_model_full

    # This is the case where we can do the merging
    if not np.all(m.spike_templates == m.spike_clusters) and m.sparse_clusters.cols is None:
        assert len(m.merge_map) > 0
        assert not np.array_equal(m.sparse_clusters.data, m.sparse_templates.data)
        assert m.sparse_clusters.data.shape[0] == m.n_clusters
        assert m.sparse_templates.data.shape[0] == m.n_templates

    else:
        assert len(m.merge_map) == 0
        assert np.array_equal(m.sparse_clusters.data, m.sparse_templates.data)
        assert np.array_equal(m.n_templates, m.n_clusters)


def test_model_save(template_model_full):
    m = template_model_full
    m.save_metadata('test', {1: 1})
    m.save_spike_clusters(m.spike_clusters)


def test_model_spike_waveforms(template_path_full):
    model = load_model(template_path_full)

    if model.traces is not None:
        traces = model.traces[:]
        assert isinstance(traces, np.ndarray)

    waveforms = {}
    for tid in model.template_ids:
        spike_ids = model.get_template_spikes(tid)
        channel_ids = model.get_template_channels(tid)
        waveforms[tid] = model.get_waveforms(spike_ids, channel_ids)

    # Export the waveforms.
    model.save_spikes_subset_waveforms(1000, 16)
    # Fill spike_waveforms after saving them.
    model.spike_waveforms = model._load_spike_waveforms()

    # Check the waveforms loaded from the spike subset waveforms arrays.
    nsw = model.n_samples_waveforms // 2
    if model.spike_waveforms is None:
        return
    for tid in model.template_ids:
        spike_ids = model.get_template_spikes(tid)
        channel_ids = model.get_template_channels(tid)
        spike_ids = np.intersect1d(spike_ids, model.spike_waveforms.spike_ids)
        w = model.get_waveforms(spike_ids, channel_ids)

        # Check the 2 ways of getting the waveforms.
        ae(w, waveforms[tid])

        if model.traces is not None:
            # Check each array with the ground truth, obtained from the raw data.
            for i, spike in enumerate(spike_ids):
                t = int(model.spike_samples[spike])
                wt = traces[t - nsw:t + nsw, channel_ids]

                ae(waveforms[tid][i], wt)
                ae(w[i], wt)

    model.close()


def test_model_metadata_1(template_model_full):
    m = template_model_full

    assert m.metadata.get('group', {}).get(4, None) == 'good'
    assert m.metadata.get('unknown', {}).get(4, None) is None

    assert m.metadata.get('quality', {}).get(6, None) is None
    m.save_metadata('quality', {6: 3})
    m.metadata = m._load_metadata()
    assert m.metadata.get('quality', {}).get(6, None) == 3


def test_model_metadata_2(template_model):
    m = template_model

    m.save_metadata('quality', {0: None, 1: 1})
    m.metadata = m._load_metadata()
    assert m.metadata.get('quality', {}).get(0, None) is None
    assert m.metadata.get('quality', {}).get(1, None) == 1


def test_model_metadata_3(template_model):
    m = template_model

    assert m.metadata.get('met1', {}).get(2, None) == 123.4
    assert m.metadata.get('met1', {}).get(3, None) == 5.678
    assert m.metadata.get('met1', {}).get(4, None) is None
    assert m.metadata.get('met1', {}).get(5, None) is None

    assert m.metadata.get('met2', {}).get(2, None) == 'hello world 1'
    assert m.metadata.get('met2', {}).get(3, None) is None
    assert m.metadata.get('met2', {}).get(4, None) is None
    assert m.metadata.get('met2', {}).get(5, None) == 'hello world 2'


def test_model_spike_attributes(template_model_full):
    model = template_model_full
    assert set(model.spike_attributes.keys()) == set(('randn', 'works'))
    assert model.spike_attributes.works.shape == (model.n_spikes,)
    assert model.spike_attributes.randn.shape == (model.n_spikes, 2)



# -*- coding: utf-8 -*-

"""Test fixtures."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging
import shutil

import numpy as np
from pytest import fixture

from phylib.utils._misc import write_text, write_tsv
from ..model import load_model, write_array
from phylib.io.datasets import download_test_file

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Fixtures
#------------------------------------------------------------------------------

_FILES = [
    'template/params.py',
    'template/sim_binary.dat',
    'template/spike_times.npy',
    'template/spike_templates.npy',
    'template/spike_clusters.npy',
    'template/amplitudes.npy',

    'template/cluster_group.tsv',

    'template/channel_map.npy',
    'template/channel_positions.npy',
    'template/channel_shanks.npy',

    'template/similar_templates.npy',
    'template/whitening_mat.npy',

    'template/templates.npy',
    'template/template_ind.npy',

    'template/pc_features.npy',
    'template/pc_feature_ind.npy',
    'template/pc_feature_spike_ids.npy',

    'template/template_features.npy',
    'template/template_feature_ind.npy',
    'template/template_feature_spike_ids.npy',
]


def _remove(path):
    if path.exists():
        path.unlink()
        logger.debug("Removed %s.", path)


def _make_dataset(tempdir, param='dense', has_spike_attributes=True):
    np.random.seed(0)

    # Download the dataset.
    paths = list(map(download_test_file, _FILES))
    # Copy the dataset to a temporary directory.
    for path in paths:
        to_path = tempdir / path.name
        # Skip sparse arrays if is_sparse is False.
        if param == 'sparse' and ('_ind.' in str(to_path) or 'spike_ids.' in str(to_path)):
            continue
        logger.debug("Copying file to %s.", to_path)
        shutil.copy(path, to_path)

    # Some changes to files if 'misc' fixture parameter.
    if param == 'misc':
        # Remove spike_clusters and recreate it from spike_templates.
        _remove(tempdir / 'spike_clusters.npy')
        # Replace spike_times.npy, in samples, by spikes.times.npy, in seconds.
        if (tempdir / 'spike_times.npy').exists():
            st = np.load(tempdir / 'spike_times.npy').squeeze()
            st_r = st + np.random.randint(low=-20000, high=+20000, size=st.size)
            assert st_r.shape == st.shape
            # Reordered spikes.
            np.save(tempdir / 'spike_times_reordered.npy', st_r)
            np.save(tempdir / 'spikes.times.npy', st / 25000.)  # sample rate
            _remove(tempdir / 'spike_times.npy')
        # Buggy TSV file should not cause a crash.
        write_text(tempdir / 'error.tsv', '')
        # Remove some non-necessary files.
        _remove(tempdir / 'template_features.npy')
        _remove(tempdir / 'pc_features.npy')
        _remove(tempdir / 'channel_probes.npy')
        _remove(tempdir / 'channel_shanks.npy')
        _remove(tempdir / 'amplitudes.npy')
        _remove(tempdir / 'whitening_mat.npy')
        _remove(tempdir / 'whitening_mat_inv.npy')
        _remove(tempdir / 'sim_binary.dat')

    if param == 'merged':
        # remove this file to make templates dense
        _remove(tempdir / 'template_ind.npy')
        clus = np.load(tempdir / 'spike_clusters.npy')
        max_clus = np.max(clus)
        # merge cluster 0 and 1
        clus[np.bitwise_or(clus == 0, clus == 1)] = max_clus + 1
        # split cluster 9 into two clusters
        idx = np.where(clus == 9)[0]
        clus[idx[0:3]] = max_clus + 2
        clus[idx[3:]] = max_clus + 3
        np.save(tempdir / 'spike_clusters.npy', clus)

    # Spike attributes.
    if has_spike_attributes:
        write_array(tempdir / 'spike_fail.npy', np.full(10, np.nan))  # wrong number of spikes
        write_array(tempdir / 'spike_works.npy', np.random.rand(314))
        write_array(tempdir / 'spike_randn.npy', np.random.randn(314, 2))

    # TSV file with cluster data.
    write_tsv(
        tempdir / 'cluster_Amplitude.tsv', [{'cluster_id': 1, 'Amplitude': 123.4}],
        first_field='cluster_id')

    write_tsv(
        tempdir / 'cluster_metrics.tsv', [
            {'cluster_id': 2, 'met1': 123.4, 'met2': 'hello world 1'},
            {'cluster_id': 3, 'met1': 5.678},
            {'cluster_id': 5, 'met2': 'hello world 2'},
        ])

    template_path = tempdir / paths[0].name
    return template_path


@fixture(scope='function', params=('dense', 'sparse', 'misc', 'merged'))
def template_path_full(tempdir, request):
    return _make_dataset(tempdir, request.param)


@fixture(scope='function')
def template_path(tempdir, request):
    return _make_dataset(tempdir, param='dense', has_spike_attributes=False)


@fixture
def template_model_full(template_path_full):
    model = load_model(template_path_full)
    yield model
    model.close()


@fixture
def template_model(template_path):
    model = load_model(template_path)
    yield model
    model.close()



# -*- coding: utf-8 -*-

"""Utility functions used for tests."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from contextlib import contextmanager
from io import StringIO
import logging
import os
import sys

from numpy.testing import assert_array_equal as ae
from numpy.testing import assert_allclose as ac

from ._types import _is_array_like

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Utility functions
#------------------------------------------------------------------------------

@contextmanager
def captured_output():
    """Context manager that captures all output to stdout and stderr."""
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


@contextmanager
def captured_logging(name=None):
    """Context manager that captures all logging."""
    logger = logging.getLogger(name)
    handlers = list(logger.handlers)
    for handler in logger.handlers:
        logger.removeHandler(handler)
    buffer = StringIO()
    handler = logging.StreamHandler(buffer)
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    yield buffer
    buffer.flush()
    logger.removeHandler(handler)
    for handler in handlers:
        logger.addHandler(handler)
    handler.close()


def _assert_equal(d_0, d_1):
    """Check that two objects are equal."""
    # Compare arrays.
    if _is_array_like(d_0):
        try:
            ae(d_0, d_1)
        except AssertionError:
            ac(d_0, d_1)
    # Compare dicts recursively.
    elif isinstance(d_0, dict):
        assert set(d_0) == set(d_1)
        for k_0 in d_0:
            _assert_equal(d_0[k_0], d_1[k_0])
    else:
        # General comparison.
        assert d_0 == d_1


def _in_travis():  # pragma: no cover
    """Return whether we're in travis."""
    return 'TRAVIS' in os.environ



# -*- coding: utf-8 -*-
from __future__ import print_function

"""Simple event system."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from contextlib import contextmanager
import logging
import string
import re
from functools import partial

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Event system
#------------------------------------------------------------------------------

class EventEmitter(object):
    """Singleton class that emits events and accepts registered callbacks.

    Example
    -------

    ```python
    class MyClass(EventEmitter):
        def f(self):
            self.emit('my_event', 1, key=2)

    o = MyClass()

    # The following function will be called when `o.f()` is called.
    @o.connect
    def on_my_event(arg, key=None):
        print(arg, key)

    ```

    """

    def __init__(self):
        self.reset()
        self.is_silent = False

    def set_silent(self, silent):
        """Set whether to silence the events."""
        self.is_silent = silent

    def reset(self):
        """Remove all registered callbacks."""
        self._callbacks = []

    def _get_on_name(self, func):
        """Return `eventname` when the function name is `on_<eventname>()`."""
        r = re.match("^on_(.+)$", func.__name__)
        if r:
            event = r.group(1)
        else:
            raise ValueError("The function name should be "
                             "`on_<eventname>`().")
        return event

    @contextmanager
    def silent(self):
        """Prevent all callbacks to be called if events are raised
        in the context manager.
        """
        self.is_silent = not(self.is_silent)
        yield
        self.is_silent = not(self.is_silent)

    def connect(self, func=None, event=None, sender=None, **kwargs):
        """Register a callback function to a given event.

        To register a callback function to the `spam` event, where `obj` is
        an instance of a class deriving from `EventEmitter`:

        ```python
        @obj.connect(sender=sender)
        def on_spam(sender, arg1, arg2):
            pass
        ```

        This is called when `obj.emit('spam', sender, arg1, arg2)` is called.

        Several callback functions can be registered for a given event.

        The registration order is conserved and may matter in applications.

        """
        if func is None:
            return partial(self.connect, event=event, sender=sender, **kwargs)

        # Get the event name from the function.
        if event is None:
            event = self._get_on_name(func)

        # We register the callback function.
        self._callbacks.append((event, sender, func, kwargs))

        return func

    def unconnect(self, *items):
        """Unconnect specified callback functions or senders."""
        self._callbacks = [
            (event, sender, f, kwargs)
            for (event, sender, f, kwargs) in self._callbacks
            if f not in items and sender not in items and
            getattr(f, '__self__', None) not in items]

    def emit(self, event, sender, *args, **kwargs):
        """Call all callback functions registered with an event.

        Any positional and keyword arguments can be passed here, and they will
        be forwarded to the callback functions.

        Return the list of callback return results.

        """
        if self.is_silent:
            return
        sender_name = sender.__class__.__name__
        logger.log(
            5, "Emit %s.%s(%s, %s)", sender_name, event,
            ', '.join(map(str, args)), ', '.join('%s=%s' % (k, v) for k, v in kwargs.items()))
        # Call the last callback if this is a single event.
        single = kwargs.pop('single', None)
        res = []
        # Put `last=True` callbacks at the end.
        callbacks = [c for c in self._callbacks if not c[-1].get('last', None)]
        callbacks += [c for c in self._callbacks if c[-1].get('last', None)]
        for e, s, f, k in callbacks:
            if e == event and (s is None or s == sender):
                f_name = getattr(f, '__qualname__', getattr(f, '__name__', str(f)))
                s_name = s.__class__.__name__
                logger.log(5, "Callback %s (%s).", f_name, s_name)
                res.append(f(sender, *args, **kwargs))
                if single:
                    return res[-1]
        return res


#------------------------------------------------------------------------------
# Progress reporter
#------------------------------------------------------------------------------

class PartialFormatter(string.Formatter):
    """Prevent KeyError when a format parameter is absent."""
    def get_field(self, field_name, args, kwargs):
        try:
            return super(PartialFormatter, self).get_field(field_name,
                                                           args,
                                                           kwargs)
        except (KeyError, AttributeError):
            return None, field_name

    def format_field(self, value, spec):
        """Format a field."""
        if value is None:
            return '?'
        try:
            return super(PartialFormatter, self).format_field(value, spec)
        except ValueError:
            return '?'


def _default_on_progress(sender, message, value, value_max, end='\r', **kwargs):
    if value_max == 0:  # pragma: no cover
        return
    if value <= value_max:
        progress = 100 * value / float(value_max)
        fmt = PartialFormatter()
        kwargs['value'] = value
        kwargs['value_max'] = value_max
        print(fmt.format(message, progress=progress, **kwargs), end=end)


def _default_on_complete(message, end='\n', **kwargs):
    # Override the initializing message and clear the terminal
    # line.
    fmt = PartialFormatter()
    print(fmt.format(message + '\033[K', **kwargs), end=end)


class ProgressReporter(object):
    """A class that reports progress done.

    Example
    -------

    ```python
    pr = ProgressReporter()
    pr.set_progress_message("Progress: {progress}%...")
    pr.set_complete_message("Completed!")
    pr.value_max = 10

    for i in range(10):
        pr.value += 1  # or pr.increment()
    ```

    You can also add custom keyword arguments in `pr.increment()`: these
    will be replaced in the message string.

    Emits
    -----

    * `progress(value, value_max)`
    * `complete()`

    """
    def __init__(self):
        super(ProgressReporter, self).__init__()
        self._value = 0
        self._value_max = 0
        self._has_completed = False

    def set_progress_message(self, message, line_break=False):
        """Set a progress message.

        The string needs to contain `{progress}`.

        """

        end = '\r' if not line_break else None

        @connect(sender=self)
        def on_progress(sender, value, value_max, **kwargs):
            kwargs['end'] = None if value == value_max else end
            _default_on_progress(sender, message, value, value_max, **kwargs)

    def set_complete_message(self, message):
        """Set a complete message."""

        @connect(sender=self)
        def on_complete(sender, **kwargs):
            _default_on_complete(message, **kwargs)

    def _set_value(self, value, **kwargs):
        if value < self._value_max:
            self._has_completed = False
        self._value = value
        emit('progress', self, self._value, self._value_max, **kwargs)
        if not self._has_completed and self._value >= self._value_max:
            emit('complete', self, **kwargs)
            self._has_completed = True

    def increment(self, **kwargs):
        """Equivalent to `self.value += 1`.

        Custom keywoard arguments can also be passed to be processed in the
        progress message format string.

        """
        self._set_value(self._value + 1, **kwargs)

    def reset(self, value_max=None):
        """Reset the value to 0 and the value max to a given value."""
        self._value = 0
        if value_max is not None:
            self._value_max = value_max

    @property
    def value(self):
        """Current value (integer)."""
        return self._value

    @value.setter
    def value(self, value):
        self._set_value(value)

    @property
    def value_max(self):
        """Maximum value (integer)."""
        return self._value_max

    @value_max.setter
    def value_max(self, value_max):
        if value_max > self._value_max:
            self._has_completed = False
        self._value_max = value_max

    def is_complete(self):
        """Return whether the task has completed."""
        return self._value >= self._value_max

    def set_complete(self, **kwargs):
        """Set the task as complete."""
        self._set_value(self.value_max, **kwargs)

    @property
    def progress(self):
        """Return the current progress as a float value in `[0, 1]`."""
        return self._value / float(self._value_max)


#------------------------------------------------------------------------------
# Global event system
#------------------------------------------------------------------------------

_EVENT = EventEmitter()

emit = _EVENT.emit
connect = _EVENT.connect
unconnect = _EVENT.unconnect
silent = _EVENT.silent
set_silent = _EVENT.set_silent
reset = _EVENT.reset



# -*- coding: utf-8 -*-

"""Utility functions for NumPy arrays."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging
from math import floor, ceil
from operator import itemgetter
from pathlib import Path

import numpy as np

from phylib.utils import _as_scalar, _as_scalars
from phylib.utils._types import _as_array

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Utility functions
#------------------------------------------------------------------------------

def _clip(x, a, b):
    return max(a, min(b, x))


def _range_from_slice(myslice, start=None, stop=None, step=None, length=None):
    """Convert a slice to an array of integers."""
    assert isinstance(myslice, slice)
    # Find 'step'.
    step = myslice.step if myslice.step is not None else step
    if step is None:
        step = 1
    # Find 'start'.
    start = myslice.start if myslice.start is not None else start
    if start is None:
        start = 0
    # Find 'stop' as a function of length if 'stop' is unspecified.
    stop = myslice.stop if myslice.stop is not None else stop
    if length is not None:
        stop_inferred = floor(start + step * length)
        if stop is not None and stop < stop_inferred:
            raise ValueError("'stop' ({stop}) and ".format(stop=stop) +
                             "'length' ({length}) ".format(length=length) +
                             "are not compatible.")
        stop = stop_inferred
    if stop is None and length is None:
        raise ValueError("'stop' and 'length' cannot be both unspecified.")
    myrange = np.arange(start, stop, step)
    # Check the length if it was specified.
    if length is not None:
        assert len(myrange) == length
    return myrange


def _unique(x):
    """Faster version of np.unique().

    This version is restricted to 1D arrays of non-negative integers.

    It is only faster if len(x) >> len(unique(x)).

    """
    if x is None or len(x) == 0:
        return np.array([], dtype=np.int64)
    # WARNING: only keep positive values.
    # cluster=-1 means "unclustered".
    x = _as_array(x)
    x = x[x >= 0]
    bc = np.bincount(x)
    return np.nonzero(bc)[0]


def _normalize(arr, keep_ratio=False):
    """Normalize an array into [0, 1]."""
    (x_min, y_min), (x_max, y_max) = arr.min(axis=0), arr.max(axis=0)

    if keep_ratio:
        a = 1. / max(x_max - x_min, y_max - y_min)
        ax = ay = a
        bx = .5 - .5 * a * (x_max + x_min)
        by = .5 - .5 * a * (y_max + y_min)
    else:
        ax = 1. / (x_max - x_min)
        ay = 1. / (y_max - y_min)
        bx = -x_min / (x_max - x_min)
        by = -y_min / (y_max - y_min)

    arr_n = arr.copy()
    arr_n[:, 0] *= ax
    arr_n[:, 0] += bx
    arr_n[:, 1] *= ay
    arr_n[:, 1] += by

    return arr_n


def _index_of(arr, lookup):
    """Replace scalars in an array by their indices in a lookup table.

    Implicitely assume that:

    * All elements of arr and lookup are non-negative integers.
    * All elements or arr belong to lookup.

    This is not checked for performance reasons.

    """
    # Equivalent of np.digitize(arr, lookup) - 1, but much faster.
    # TODO: assertions to disable in production for performance reasons.
    # TODO: np.searchsorted(lookup, arr) is faster on small arrays with large
    # values
    lookup = np.asarray(lookup, dtype=np.int32)
    m = (lookup.max() if len(lookup) else 0) + 1
    tmp = np.zeros(m + 1, dtype=int)
    # Ensure that -1 values are kept.
    tmp[-1] = -1
    if len(lookup):
        tmp[lookup] = np.arange(len(lookup))
    return tmp[arr]


def _pad(arr, n, dir='right'):
    """Pad an array with zeros along the first axis.

    Parameters
    ----------

    n : int
        Size of the returned array in the first axis.
    dir : str
        Direction of the padding. Must be one 'left' or 'right'.

    """
    assert dir in ('left', 'right')
    if n < 0:
        raise ValueError("'n' must be positive: {0}.".format(n))
    elif n == 0:
        return np.zeros((0,) + arr.shape[1:], dtype=arr.dtype)
    n_arr = arr.shape[0]
    shape = (n,) + arr.shape[1:]
    if n_arr == n:
        assert arr.shape == shape
        return arr
    elif n_arr < n:
        out = np.zeros(shape, dtype=arr.dtype)
        if dir == 'left':
            out[-n_arr:, ...] = arr
        elif dir == 'right':
            out[:n_arr, ...] = arr
        assert out.shape == shape
        return out
    else:
        if dir == 'left':
            out = arr[-n:, ...]
        elif dir == 'right':
            out = arr[:n, ...]
        assert out.shape == shape
        return out


def _get_padded(data, start, end):
    """Return `data[start:end]` filling in with zeros outside array bounds

    Assumes that either `start<0` or `end>len(data)` but not both.

    """
    if start < 0 and end > data.shape[0]:
        raise RuntimeError()
    if start < 0:
        start_zeros = np.zeros((-start, data.shape[1]),
                               dtype=data.dtype)
        return np.vstack((start_zeros, data[:end]))
    elif end > data.shape[0]:
        end_zeros = np.zeros((end - data.shape[0], data.shape[1]),
                             dtype=data.dtype)
        return np.vstack((data[start:], end_zeros))
    else:
        return data[start:end]


def _get_data_lim(arr, n_spikes=None):
    n = arr.shape[0]
    k = max(1, n // n_spikes) if n_spikes else 1
    arr = np.abs(arr[::k])
    n = arr.shape[0]
    arr = arr.reshape((n, -1))
    return arr.max() or 1.


def get_closest_clusters(cluster_id, cluster_ids, sim_func, max_n=None):
    """Return a list of pairs `(cluster, similarity)` sorted by decreasing
    similarity to a given cluster."""
    l = [(_as_scalar(candidate), _as_scalar(sim_func(cluster_id, candidate)))
         for candidate in _as_scalars(cluster_ids)]
    l = sorted(l, key=itemgetter(1), reverse=True)
    max_n = None or len(l)
    return l[:max_n]


def _flatten(l):
    return [item for sublist in l for item in sublist]


# -----------------------------------------------------------------------------
# I/O functions
# -----------------------------------------------------------------------------

def read_array(path, mmap_mode=None):
    """Read a .npy array."""
    path = Path(path)
    file_ext = path.suffix
    if file_ext == '.npy':
        return np.load(str(path), mmap_mode=mmap_mode)
    raise NotImplementedError("The file extension `{}` is not currently supported." % file_ext)


def write_array(path, arr):
    """Write an array to a .npy file."""
    path = Path(path)
    file_ext = path.suffix
    if file_ext == '.npy':
        return np.save(str(path), arr)
    raise NotImplementedError("The file extension `{}` is not currently supported." % file_ext)


# -----------------------------------------------------------------------------
# Chunking functions
# -----------------------------------------------------------------------------

def _excerpt_step(n_samples, n_excerpts=None, excerpt_size=None):
    """Compute the step of an excerpt set as a function of the number
    of excerpts or their sizes."""
    assert n_excerpts >= 2
    step = max((n_samples - excerpt_size) // (n_excerpts - 1),
               excerpt_size)
    return step


def chunk_bounds(n_samples, chunk_size, overlap=0):
    """Return chunk bounds.

    Chunks have the form:

        [ overlap/2 | chunk_size-overlap | overlap/2 ]
        s_start   keep_start           keep_end     s_end

    Except for the first and last chunks which do not have a left/right
    overlap.

    This generator yields (s_start, s_end, keep_start, keep_end).

    """
    s_start = 0
    s_end = chunk_size
    keep_start = s_start
    keep_end = s_end - overlap // 2
    yield s_start, s_end, keep_start, keep_end

    while s_end - overlap + chunk_size < n_samples:
        s_start = s_end - overlap
        s_end = s_start + chunk_size
        keep_start = keep_end
        keep_end = s_end - overlap // 2
        if s_start < s_end:
            yield s_start, s_end, keep_start, keep_end

    s_start = s_end - overlap
    s_end = n_samples
    keep_start = keep_end
    keep_end = s_end
    if s_start < s_end:
        yield s_start, s_end, keep_start, keep_end


def excerpts(n_samples, n_excerpts=None, excerpt_size=None):
    """Yield (start, end) where start is included and end is excluded."""
    assert n_excerpts >= 2
    step = _excerpt_step(n_samples, n_excerpts=n_excerpts, excerpt_size=excerpt_size)
    for i in range(n_excerpts):
        start = i * step
        if start >= n_samples:
            break
        end = min(start + excerpt_size, n_samples)
        yield start, end


def data_chunk(data, chunk, with_overlap=False):
    """Get a data chunk."""
    assert isinstance(chunk, tuple)
    if len(chunk) == 2:
        i, j = chunk
    elif len(chunk) == 4:
        if with_overlap:
            i, j = chunk[:2]
        else:
            i, j = chunk[2:]
    else:
        raise ValueError("'chunk' should have 2 or 4 elements, not {0:d}".format(len(chunk)))
    return data[i:j, ...]


def get_excerpts(data, n_excerpts=None, excerpt_size=None):
    """Return excerpts of a data array."""
    assert n_excerpts is not None
    assert excerpt_size is not None
    if len(data) < n_excerpts * excerpt_size:
        return data
    elif n_excerpts == 0:
        return data[:0]
    elif n_excerpts == 1:
        return data[:excerpt_size]
    out = np.concatenate([
        data_chunk(data, chunk)
        for chunk in excerpts(len(data), n_excerpts=n_excerpts, excerpt_size=excerpt_size)])
    assert len(out) <= n_excerpts * excerpt_size
    return out


# -----------------------------------------------------------------------------
# Spike clusters utility functions
# -----------------------------------------------------------------------------

def _spikes_in_clusters(spike_clusters, clusters):
    """Return the ids of all spikes belonging to the specified clusters."""
    if len(spike_clusters) == 0 or len(clusters) == 0:
        return np.array([], dtype=int)
    return np.nonzero(np.isin(spike_clusters, clusters))[0]


def _spikes_per_cluster(spike_clusters, spike_ids=None):
    """Return a dictionary {cluster: list_of_spikes}."""
    if spike_clusters is None or not len(spike_clusters):
        return {}
    if spike_ids is None:
        spike_ids = np.arange(len(spike_clusters)).astype(np.int64)
    # NOTE: this sort method is stable, so spike ids are increasing
    # among any cluster. Therefore we don't have to sort again down here,
    # when creating the spikes_in_clusters dictionary.
    rel_spikes = np.argsort(spike_clusters, kind='mergesort')
    abs_spikes = spike_ids[rel_spikes]
    spike_clusters = spike_clusters[rel_spikes]

    diff = np.empty_like(spike_clusters)
    diff[0] = 1
    diff[1:] = np.diff(spike_clusters)

    idx = np.nonzero(diff > 0)[0]
    clusters = spike_clusters[idx]

    # NOTE: we don't have to sort abs_spikes[...] here because the argsort
    # using 'mergesort' above is stable.
    spikes_in_clusters = {
        clusters[i]: abs_spikes[idx[i]:idx[i + 1]] for i in range(len(clusters) - 1)}
    spikes_in_clusters[clusters[-1]] = abs_spikes[idx[-1]:]

    return spikes_in_clusters


def _flatten_per_cluster(per_cluster):
    """Convert a dictionary {cluster: spikes} to a spikes array."""
    return np.unique(np.concatenate(list(per_cluster.values()))).astype(np.int64)


def grouped_mean(arr, spike_clusters):
    """Compute the mean of a spike-dependent quantity for every cluster.

    The two arguments should be 1D array with `n_spikes` elements.

    The output is a 1D array with `n_clusters` elements. The clusters are
    sorted in increasing order.

    """
    arr = np.asarray(arr)
    spike_clusters = np.asarray(spike_clusters)
    assert arr.shape[0] == len(spike_clusters)
    cluster_ids = _unique(spike_clusters)
    spike_clusters_rel = _index_of(spike_clusters, cluster_ids)
    spike_counts = np.bincount(spike_clusters_rel)
    assert len(spike_counts) == len(cluster_ids)
    t = np.zeros((len(cluster_ids),) + arr.shape[1:])
    # Compute the sum with possible repetitions.
    np.add.at(t, spike_clusters_rel, arr)
    return t / spike_counts.reshape((-1,) + (1,) * (arr.ndim - 1))


# -----------------------------------------------------------------------------
# Spike selection
# -----------------------------------------------------------------------------

def _times_in_chunks(times, chunks_kept):
    """Return the indices of the times that belong to a list of kept chunks."""
    ind = np.searchsorted(chunks_kept, times, side='right')
    return ind % 2 == 1


class SpikeSelector(object):
    """Select a given number of spikes per cluster among a subset of the chunks."""
    def __init__(
            self, get_spikes_per_cluster=None, spike_times=None,
            chunk_bounds=None, n_chunks_kept=None):
        self.get_spikes_per_cluster = get_spikes_per_cluster
        self.spike_times = spike_times
        self.chunks_kept = []
        n_chunks = len(chunk_bounds) - 1

        for i in range(0, n_chunks, max(1, int(ceil(n_chunks / n_chunks_kept)))):
            self.chunks_kept.extend(chunk_bounds[i:i + 2])
        self.chunks_kept = np.array(self.chunks_kept)

    def __call__(self, n_spk_clu, cluster_ids, subset_chunks=False, subset_spikes=None):
        """Select about n_spk_clu random spikes from each of the requested clusters, only
        in the kept chunks."""
        if not len(cluster_ids):
            return np.array([], dtype=np.int64)
        # Start with all spikes from each cluster.
        selection = {}
        for cluster in cluster_ids:
            # Get all spikes from that cluster.
            spike_ids = self.get_spikes_per_cluster(cluster)
            # Get the spike times.
            t = self.spike_times[spike_ids]
            # Keep the spikes belonging to the chunks.
            if subset_chunks:
                spike_ids = spike_ids[_times_in_chunks(t, self.chunks_kept)]
            # Keep spikes from a given subset.
            if subset_spikes is not None:
                spike_ids = np.intersect1d(spike_ids, subset_spikes)
            # Make a subselection if needed.
            if n_spk_clu is not None and n_spk_clu > 0 and len(spike_ids) > n_spk_clu:
                spike_ids = np.random.choice(spike_ids, n_spk_clu, replace=False)
            selection[cluster] = spike_ids
        # Return the concatenation of all spikes.
        return _flatten_per_cluster(selection)



# -*- coding: utf-8 -*-

"""Test probe merging."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np

from ..merge import Merger
from phylib.io.alf import EphysAlfCreator
from phylib.io.model import load_model
from phylib.io.tests.conftest import _make_dataset


#------------------------------------------------------------------------------
# Merging tests
#------------------------------------------------------------------------------

def test_probe_merge_1(tempdir):
    out_dir = tempdir / 'merged'

    # Create two identical datasets.
    probe_names = ('probe_left', 'probe_right')
    for name in probe_names:
        (tempdir / name).mkdir(exist_ok=True, parents=True)
        _make_dataset(tempdir / name, param='dense', has_spike_attributes=False)

    subdirs = [tempdir / name for name in probe_names]

    # Merge them.
    m = Merger(subdirs, out_dir)
    single = load_model(tempdir / probe_names[0] / 'params.py')

    # Test the merged dataset.
    merged = m.merge()
    for name in ('n_spikes', 'n_channels', 'n_templates'):
        assert getattr(merged, name) == getattr(single, name) * 2
    assert merged.sample_rate == single.sample_rate


def test_probe_merge_2(tempdir):
    out_dir = tempdir / 'merged'

    # Create two identical datasets.
    probe_names = ('probe_left', 'probe_right')
    for name in probe_names:
        (tempdir / name).mkdir(exist_ok=True, parents=True)
        _make_dataset(tempdir / name, param='dense', has_spike_attributes=False)
    subdirs = [tempdir / name for name in probe_names]

    # Add small shift in the spike times of the second probe.
    single = load_model(tempdir / probe_names[0] / 'params.py')
    st_path = tempdir / 'probe_right/spike_times.npy'
    np.save(st_path, single.spike_samples + 1)
    # make amplitudes unique and growing so they can serve as key and sorting indices
    single.amplitudes = np.linspace(5, 15, single.n_spikes)
    # single.spike_clusters[single.spike_clusters == 0] = 12
    for m, subdir in enumerate(subdirs):
        np.save(subdir / 'amplitudes.npy', single.amplitudes + 20 * m)
        np.save(subdir / 'spike_clusters.npy', single.spike_clusters)

    # Merge them.
    m = Merger(subdirs, out_dir)
    merged = m.merge()

    # Test the merged dataset.
    for name in ('n_spikes', 'n_channels', 'n_templates'):
        assert getattr(merged, name) == getattr(single, name) * 2
    assert merged.sample_rate == single.sample_rate

    # Check the spikes.
    single = load_model(tempdir / probe_names[0] / 'params.py')

    def test_merged_single(merged, merged_original_amps=None):
        if merged_original_amps is None:
            merged_original_amps = merged.amplitudes
        _, im1, i1 = np.intersect1d(merged_original_amps, single.amplitudes, return_indices=True)
        _, im2, i2 = np.intersect1d(merged_original_amps, single.amplitudes + 20,
                                    return_indices=True)
        # intersection spans the full vector
        assert i1.size + i2.size == merged.amplitudes.size
        # test spikes
        assert np.allclose(merged.spike_times[im1], single.spike_times[i1])
        assert np.allclose(merged.spike_times[im2], single.spike_times[i2] + 4e-5)
        # test clusters
        assert np.allclose(merged.spike_clusters[im2], single.spike_clusters[i2] + 64)
        assert np.allclose(merged.spike_clusters[im1], single.spike_clusters[i1])
        # test templates
        assert np.all(merged.spike_templates[im1] - single.spike_templates[i1] == 0)
        assert np.all(merged.spike_templates[im2] - single.spike_templates[i2] == 64)
        # test probes
        assert np.all(merged.channel_probes == np.r_[single.channel_probes,
                                                     single.channel_probes + 1])
        assert np.all(merged.templates_channels[merged.templates_probes == 0] < single.n_channels)
        assert np.all(merged.templates_channels[merged.templates_probes == 1] >= single.n_channels)
        spike_probes = merged.templates_probes[merged.spike_templates]

        assert np.all(merged_original_amps[spike_probes == 0] <= 15)
        assert np.all(merged_original_amps[spike_probes == 1] >= 20)

        # np.all(merged.sparse_templates.data[:64, :, 0:32] == single.sparse_templates.data)

    # Convert into ALF and load.
    alf = EphysAlfCreator(merged).convert(tempdir / 'alf')
    test_merged_single(merged)
    test_merged_single(alf, merged_original_amps=merged.amplitudes)

    # specific test channel ids only for ALF merge dataset: the raw indices are still individual
    # file indices, the merged channel mapping is in `channels._phy_ids.npy`
    chid = np.load(tempdir.joinpath('alf', 'channels.rawInd.npy'))
    assert np.all(chid == np.r_[single.channel_mapping, single.channel_mapping])

    out_files = list(tempdir.joinpath('alf').glob('*.*'))
    cl_shape = [np.load(f).shape[0] for f in out_files if f.name.startswith('clusters.') and
                f.name.endswith('.npy')]
    sp_shape = [np.load(f).shape[0] for f in out_files if f.name.startswith('spikes.')]
    ch_shape = [np.load(f).shape[0] for f in out_files if f.name.startswith('channels.')]
    assert len(set(cl_shape)) == 1
    assert len(set(sp_shape)) == 1
    assert len(set(ch_shape)) == 1



# -*- coding: utf-8 -*-

"""Tests of misc type utility functions."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from pytest import raises

from .._types import (Bunch, _bunchify, _is_integer, _is_list, _is_float,
                      _as_list, _is_array_like, _as_array, _as_tuple,
                      _as_scalar, _as_scalars,
                      )


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def test_bunch():
    obj = Bunch()
    obj['a'] = 1
    assert obj.a == 1
    obj.b = 2
    assert obj['b'] == 2
    assert obj.copy() == obj


def test_bunchify():
    d = {'a': {'b': 0}}
    b = _bunchify(d)
    assert isinstance(b, Bunch)
    assert isinstance(b['a'], Bunch)


def test_number():
    assert not _is_integer(None)
    assert not _is_integer(3.)
    assert _is_integer(3)
    assert _is_integer(np.arange(1)[0])

    assert not _is_float(None)
    assert not _is_float(3)
    assert not _is_float(np.array([3])[0])
    assert _is_float(3.)
    assert _is_float(np.array([3.])[0])


def test_list():
    assert not _is_list(None)
    assert not _is_list(())
    assert _is_list([])

    assert _as_list(None) is None
    assert _as_list(3) == [3]
    assert _as_list([3]) == [3]
    assert _as_list((3,)) == [3]
    assert _as_list('3') == ['3']
    assert np.all(_as_list(np.array([3])) == np.array([3]))


def test_as_tuple():
    assert _as_tuple(3) == (3,)
    assert _as_tuple((3,)) == (3,)
    assert _as_tuple(None) is None
    assert _as_tuple((None,)) == (None,)
    assert _as_tuple((3, 4)) == (3, 4)
    assert _as_tuple([3]) == ([3], )
    assert _as_tuple([3, 4]) == ([3, 4], )


def test_as_scalar():
    assert _as_scalar(1) == 1
    assert _as_scalar(np.ones(1)[0]) == 1.
    assert type(_as_scalar(np.ones(1)[0])) == float

    assert _as_scalars(np.arange(3)) == [0, 1, 2]


def test_array():
    def _check(arr):
        assert isinstance(arr, np.ndarray)
        assert np.all(arr == [3])

    _check(_as_array(3))
    _check(_as_array(3.))
    _check(_as_array([3]))

    _check(_as_array(3, float))
    _check(_as_array(3., float))
    _check(_as_array([3], float))
    _check(_as_array(np.array([3])))
    with raises(ValueError):
        _check(_as_array(np.array([3]), dtype=object))
    _check(_as_array(np.array([3]), float))

    assert _as_array(None) is None
    assert not _is_array_like(None)
    assert not _is_array_like(3)
    assert _is_array_like([3])
    assert _is_array_like(np.array([3]))



# -*- coding: utf-8 -*-
# flake8: noqa

"""Utilities."""

from ._misc import (
    load_json, save_json, load_pickle, save_pickle, _fullname, read_python,
    read_text, write_text, read_tsv, write_tsv)
from ._types import (
    _is_array_like, _as_array, _as_tuple, _as_list, _as_scalar, _as_scalars,
    Bunch, _is_list, _bunchify)
from .event import ProgressReporter, emit, connect, unconnect, silent, reset, set_silent



# -*- coding: utf-8 -*-

"""Test geometry utilities."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_array_equal as ae
from numpy.testing import assert_allclose as ac
import pytest

from ..geometry import (
    linear_positions,
    staggered_positions,
    _get_data_bounds,
    _boxes_overlap,
    _binary_search,
    _find_box_size,
    get_non_overlapping_boxes,
    get_closest_box,
)


#------------------------------------------------------------------------------
# Test utilities
#------------------------------------------------------------------------------

def test_get_data_bounds():
    ae(_get_data_bounds(None), [[-1., -1., 1., 1.]])

    db0 = np.array([[0, 1, 4, 5],
                    [0, 1, 4, 5],
                    [0, 1, 4, 5]])
    arr = np.arange(6).reshape((3, 2))
    assert np.all(_get_data_bounds(None, arr) == [[0, 1, 4, 5]])

    db = db0.copy()
    assert np.all(_get_data_bounds(db, arr) == [[0, 1, 4, 5]])

    db = db0.copy()
    db[2, :] = [1, 1, 1, 1]
    assert np.all(_get_data_bounds(db, arr)[:2, :] == [[0, 1, 4, 5]])
    assert np.all(_get_data_bounds(db, arr)[2, :] == [0, 0, 2, 2])

    db = db0.copy()
    db[:2, :] = [1, 1, 1, 1]
    assert np.all(_get_data_bounds(db, arr)[:2, :] == [[0, 0, 2, 2]])
    assert np.all(_get_data_bounds(db, arr)[2, :] == [0, 1, 4, 5])


def test_boxes_overlap():

    def _get_args(boxes):
        x0, y0, x1, y1 = np.array(boxes).T
        x0 = x0[:, np.newaxis]
        x1 = x1[:, np.newaxis]
        y0 = y0[:, np.newaxis]
        y1 = y1[:, np.newaxis]
        return x0, y0, x1, y1

    boxes = [[-1, -1, 0, 0], [0.01, 0.01, 1, 1]]
    x0, y0, x1, y1 = _get_args(boxes)
    assert not _boxes_overlap(x0, y0, x1, y1)

    boxes = [[-1, -1, 0.1, 0.1], [0, 0, 1, 1]]
    x0, y0, x1, y1 = _get_args(boxes)
    assert _boxes_overlap(x0, y0, x1, y1)

    x = np.zeros((5, 1))
    x0 = x - .1
    x1 = x + .1
    y = np.linspace(-1, 1, 5)[:, np.newaxis]
    y0 = y - .2
    y1 = y + .2
    assert not _boxes_overlap(x0, y0, x1, y1)


def test_binary_search():
    def f(x):
        return x < .4
    ac(_binary_search(f, 0, 1), .4)
    ac(_binary_search(f, 0, .3), .3)
    ac(_binary_search(f, .5, 1), .5)


def test_find_box_size():
    x = np.zeros(5)
    y = np.linspace(-1, 1, 5)
    w, h = _find_box_size(x, y, margin=0)
    ac(w, .5, atol=1e-8)
    ac(h, .25, atol=1e-8)


@pytest.mark.parametrize('n_channels', [5, 500])
def test_get_non_overlapping_boxes_1(n_channels):
    x = np.zeros(n_channels)
    y = np.linspace(-1, 1, n_channels)
    box_pos, box_size = get_non_overlapping_boxes(np.c_[x, y])
    ac(box_pos[:, 0], 0, atol=1e-8)
    ac(box_pos[:, 1], -box_pos[::-1, 1], atol=1e-8)

    assert box_size[0] >= .8

    s = np.array([box_size])
    box_bounds = np.c_[box_pos - s, box_pos + s]
    assert box_bounds.min() >= -1
    assert box_bounds.max() <= +1


def test_get_non_overlapping_boxes_2():
    pos = staggered_positions(32)
    box_pos, box_size = get_non_overlapping_boxes(pos)
    assert box_size[0] >= .05

    s = np.array([box_size])
    box_bounds = np.c_[box_pos - s, box_pos + s]
    assert box_bounds.min() >= -1
    assert box_bounds.max() <= +1


def test_get_closest_box():
    n = 10
    px = np.zeros(n)
    py = np.linspace(-1, 1, n)
    box_pos = np.c_[px, py]
    w, h = (1, .9 / n)
    expected = []
    for x in (0, -1, 1, -2, +2):
        for i in range(n):
            expected.extend([
                (x, py[i], i),
                (x, py[i] - h, i),
                (x, py[i] + h, i),
                (x, py[i] - 1.25 * h, max(0, min(i - 1, n - 1))),
                (x, py[i] + 1.25 * h, max(0, min(i + 1, n - 1))),
            ])
    for x, y, i in expected:
        assert get_closest_box((x, y), box_pos, (w, h)) == i


def test_positions():
    probe = staggered_positions(31)
    assert probe.shape == (31, 2)
    ae(probe[-1], (0, 0))

    probe = linear_positions(29)
    assert probe.shape == (29, 2)



# -*- coding: utf-8 -*-

"""Mock datasets."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
import numpy.random as nr


#------------------------------------------------------------------------------
# Artificial data
#------------------------------------------------------------------------------

def artificial_waveforms(n_spikes=None, n_samples=None, n_channels=None):
    return .25 * nr.normal(size=(n_spikes, n_samples, n_channels))


def artificial_features(*args):
    return .25 * nr.normal(size=args)


def artificial_masks(n_spikes=None, n_channels=None):
    masks = nr.uniform(size=(n_spikes, n_channels))
    masks[masks < .25] = 0
    return masks


def artificial_traces(n_samples, n_channels):
    return .25 * nr.normal(size=(n_samples, n_channels))


def artificial_spike_clusters(n_spikes, n_clusters, low=0):
    return nr.randint(size=n_spikes, low=low, high=max(1, n_clusters))


def artificial_spike_samples(n_spikes, max_isi=50):
    return np.cumsum(nr.randint(low=0, high=max_isi, size=n_spikes))


def artificial_correlograms(n_clusters, n_samples):
    return nr.uniform(size=(n_clusters, n_clusters, n_samples))



# -*- coding: utf-8 -*-

"""Utility functions."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np


#------------------------------------------------------------------------------
# Various Python utility functions
#------------------------------------------------------------------------------

_ACCEPTED_ARRAY_DTYPES = (
    float, np.float32, np.float64, int, np.int8, np.int16, np.uint8, np.uint16,
    np.int32, np.int64, np.uint32, np.uint64, bool)


class Bunch(dict):
    """A subclass of dictionary with an additional dot syntax."""
    def __init__(self, *args, **kwargs):
        super(Bunch, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def copy(self):
        """Return a new Bunch instance which is a copy of the current Bunch instance."""
        return Bunch(super(Bunch, self).copy())


def _bunchify(b):
    """Ensure all dict elements are Bunch."""
    assert isinstance(b, dict)
    b = Bunch(b)
    for k in b:
        if isinstance(b[k], dict):
            b[k] = Bunch(b[k])
    return b


def _is_list(obj):
    """Return whether an object is a list."""
    return isinstance(obj, list)


def _as_scalar(obj):
    """Return whether an object is a scalar number (integer or floating point number)."""
    if isinstance(obj, np.generic):
        return obj.item()
    assert isinstance(obj, (int, float))
    return obj


def _as_scalars(arr):
    """Make sure a list only contains scalar numbers."""
    return [_as_scalar(x) for x in arr]


def _is_integer(x):
    """Return whether an object is an integer."""
    return isinstance(x, (int, np.generic))


def _is_float(x):
    """Return whether an object is a floating point number."""
    return isinstance(x, (float, np.float32, np.float64))


def _as_list(obj):
    """Ensure an object is a list."""
    if obj is None:
        return None
    elif isinstance(obj, str):
        return [obj]
    elif isinstance(obj, tuple):
        return list(obj)
    elif not hasattr(obj, '__len__'):
        return [obj]
    else:
        return obj


def _is_array_like(arr):
    """Return whether an object is an array or a list."""
    return isinstance(arr, (list, np.ndarray))


def _as_array(arr, dtype=None):
    """Convert an object to a numerical NumPy array.

    Avoid a copy if possible.

    """
    if arr is None:
        return None
    if isinstance(arr, np.ndarray) and dtype is None:
        return arr
    if isinstance(arr, (int, float)):
        arr = [arr]
    out = np.asarray(arr)
    if dtype is not None:
        if out.dtype != dtype:
            out = out.astype(dtype)
    if out.dtype not in _ACCEPTED_ARRAY_DTYPES:
        raise ValueError("'arr' seems to have an invalid dtype: "
                         "{0:s}".format(str(out.dtype)))
    return out


def _as_tuple(item):
    """Ensure an item is a tuple."""
    if item is None:
        return None
    elif not isinstance(item, tuple):
        return (item,)
    else:
        return item



# -*- coding: utf-8 -*-

"""Test event system."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from pytest import raises

from ..event import EventEmitter, ProgressReporter, connect


#------------------------------------------------------------------------------
# Test event system
#------------------------------------------------------------------------------

def test_event_system():
    ev = EventEmitter()

    _list = []

    with raises(ValueError):
        ev.connect(lambda x: x)

    @ev.connect
    def on_my_event(sender, arg, kwarg=None):
        _list.append((arg, kwarg))

    ev.emit('my_event', ev, 'a')
    assert _list == [('a', None)]

    ev.emit('my_event', ev, 'b', 'c')
    assert _list == [('a', None), ('b', 'c')]

    ev.unconnect(on_my_event)

    ev.emit('my_event', ev, 'b', 'c')
    assert _list == [('a', None), ('b', 'c')]


def test_event_silent():
    ev = EventEmitter()

    _list = []

    @ev.connect()
    def on_test(sender, x):
        _list.append(x)

    ev.emit('test', ev, 1)
    assert _list == [1]

    with ev.silent():
        ev.emit('test', ev, 1)
    assert _list == [1]

    ev.set_silent(True)


def test_event_single():
    ev = EventEmitter()

    l = []

    @ev.connect(event='test')
    def on_test_bou(sender):
        l.append(0)

    @ev.connect  # noqa
    def on_test(sender):
        l.append(1)

    ev.emit('test', ev)
    assert l == [0, 1]

    ev.emit('test', ev, single=True)
    assert l == [0, 1, 0]


#------------------------------------------------------------------------------
# Test progress reporter
#------------------------------------------------------------------------------

def test_progress_reporter():
    """Test the progress reporter."""
    pr = ProgressReporter()

    _reported = []
    _completed = []

    @connect(sender=pr)
    def on_progress(sender, value, value_max):
        # value is the sum of the values, value_max the sum of the max values
        _reported.append((value, value_max))

    @connect(sender=pr)
    def on_complete(sender):
        _completed.append(True)

    pr.value_max = 10
    pr.value = 0
    pr.value = 5
    assert pr.value == 5
    assert pr.progress == .5
    assert not pr.is_complete()
    pr.value = 10
    assert pr.is_complete()
    assert pr.progress == 1.
    assert _completed == [True]

    pr.value_max = 11
    assert not pr.is_complete()
    assert pr.progress < 1.
    pr.set_complete()
    assert pr.is_complete()
    assert pr.progress == 1.

    assert _reported == [(0, 10), (5, 10), (10, 10), (11, 11)]
    assert _completed == [True, True]

    pr.value = 10
    # Only trigger a complete event once.
    pr.value = pr.value_max
    pr.value = pr.value_max
    assert _completed == [True, True, True]


def test_progress_message():
    """Test messages with the progress reporter."""
    pr = ProgressReporter()
    pr.reset(5)
    pr.set_progress_message("The progress is {progress}%. ({hello:d})")
    pr.set_complete_message("Finished {hello}.")

    pr.value_max = 10
    pr.value = 0
    print()
    pr.value = 5
    print()
    pr.increment()
    print()
    pr.increment(hello='hello')
    print()
    pr.increment(hello=3)
    print()
    pr.value = 10



# -*- coding: utf-8 -*-

"""Utility functions."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import base64
import csv
from importlib import import_module
import json
import logging
import os
from pathlib import Path
import subprocess
from textwrap import dedent

import numpy as np

from ._types import _is_integer

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# JSON utility functions
#------------------------------------------------------------------------------

def _encode_qbytearray(arr):
    """Encode binary arrays with base64."""
    b = arr.toBase64().data()
    data_b64 = base64.b64encode(b).decode('utf8')
    return data_b64


def _decode_qbytearray(data_b64):
    """Decode binary arrays with base64."""
    encoded = base64.b64decode(data_b64)
    try:
        from PyQt5.QtCore import QByteArray
        out = QByteArray.fromBase64(encoded)
    except ImportError:  # pragma: no cover
        pass
    return out


class _CustomEncoder(json.JSONEncoder):
    """JSON encoder that accepts NumPy arrays."""
    def default(self, obj):
        if isinstance(obj, np.ndarray) and obj.ndim == 1 and obj.shape[0] <= 10:
            # Serialize small arrays in clear text (lists of numbers).
            return obj.tolist()
        elif isinstance(obj, np.ndarray):
            obj_contiguous = np.ascontiguousarray(obj)
            data_b64 = base64.b64encode(obj_contiguous.data).decode('utf8')
            return dict(__ndarray__=data_b64, dtype=str(obj.dtype), shape=obj.shape)
        elif obj.__class__.__name__ == 'QByteArray':
            return {'__qbytearray__': _encode_qbytearray(obj)}
        elif isinstance(obj, np.generic):
            return obj.item()
        return super(_CustomEncoder, self).default(obj)  # pragma: no cover


def _json_custom_hook(d):
    """Serialize NumPy arrays."""
    if isinstance(d, dict) and '__ndarray__' in d:
        data = base64.b64decode(d['__ndarray__'])
        return np.frombuffer(data, d['dtype']).reshape(d['shape'])
    elif isinstance(d, dict) and '__qbytearray__' in d:
        return _decode_qbytearray(d['__qbytearray__'])
    return d


def _intify_keys(d):
    """Make sure all integer strings in a dictionary are converted into integers."""
    assert isinstance(d, dict)
    out = {}
    for k, v in d.items():
        if isinstance(k, str) and k.isdigit():
            k = int(k)
        out[k] = v
    return out


def _stringify_keys(d):
    """Make sure all integers in a dictionary are converted into strings."""
    assert isinstance(d, dict)
    out = {}
    for k, v in d.items():
        if _is_integer(k):
            k = str(k)
        out[k] = v
    return out


def _pretty_floats(obj, n=2):
    """Display floating point numbers properly."""
    if isinstance(obj, (float, np.float64, np.float32)):
        return ('%.' + str(n) + 'f') % obj
    elif isinstance(obj, dict):
        return dict((k, _pretty_floats(v)) for k, v in obj.items())
    elif isinstance(obj, (list, tuple)):
        return list(map(_pretty_floats, obj))
    return obj


def load_json(path):
    """Load a JSON file."""
    path = Path(path)
    if not path.exists():
        raise IOError("The JSON file `{}` doesn't exist.".format(path))
    contents = path.read_text()
    if not contents:
        return {}
    out = json.loads(contents, object_hook=_json_custom_hook)
    return _intify_keys(out)


def save_json(path, data):
    """Save a dictionary to a JSON file.

    Support NumPy arrays and QByteArray objects. NumPy arrays are saved as base64-encoded strings,
    except for 1D arrays with less than 10 elements, which are saved as a list for human
    readability.

    """
    assert isinstance(data, dict)
    data = _stringify_keys(data)
    path = Path(path)
    ensure_dir_exists(path.parent)
    with path.open('w') as f:
        json.dump(data, f, cls=_CustomEncoder, indent=2, sort_keys=True)


#------------------------------------------------------------------------------
# Other read/write functions
#------------------------------------------------------------------------------

def load_pickle(path):
    """Load a pickle file using joblib."""
    from joblib import load
    return load(path)


def save_pickle(path, data):
    """Save data to a pickle file using joblib."""
    from joblib import dump
    return dump(data, path)


def read_python(path):
    """Read a Python file.

    Parameters
    ----------

    path : str or Path

    Returns
    -------

    metadata : dict
        A dictionary containing all variables defined in the Python file (with `exec()`).

    """
    path = Path(path)
    if not path.exists():  # pragma: no cover
        raise IOError("Path %s does not exist.", path)
    contents = path.read_text()
    metadata = {}
    exec(contents, {}, metadata)
    metadata = {k.lower(): v for (k, v) in metadata.items()}
    return metadata


def write_python(path, data):
    """Write a dictionary in a Python file.

    Parameters
    ----------

    path : str or Path
        Path to the Python file to write.
    data : dict
        A key-value mapping to write as a Python file.

    Returns
    -------

    """
    with open(path, 'w') as f:
        for k, v in data.items():
            if isinstance(v, str):
                v = '"%s"' % v
            f.write('%s = %s\n' % (k, str(v)))


def read_text(path):
    """Read a text file."""
    path = Path(path)
    return path.read_text()


def write_text(path, contents):
    """Write a text file."""
    contents = dedent(contents)
    path = Path(path)
    ensure_dir_exists(path.parent)
    path.write_text(contents)


def _try_make_number(value):
    """Convert a string into an int or float if possible, otherwise do nothing."""
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            return value
    raise ValueError()


def read_tsv(path):
    """Read a CSV/TSV file.

    Returns
    -------

    data : list of dicts

    """
    path = Path(path)
    data = []
    if not path.exists():
        logger.debug("%s does not exist, skipping.", path)
        return data
    # Find whether the delimiter is tab or comma.
    with path.open('r') as f:
        delimiter = '\t' if '\t' in f.readline() else ','
    with path.open('r') as f:
        reader = csv.reader(f, delimiter=delimiter)
        # Skip the header.
        field_names = list(next(reader))
        for row in reader:
            data.append({k: _try_make_number(v) for k, v in zip(field_names, row) if v != ''})
    logger.log(5, "Read %s.", path)
    return data


def write_tsv(path, data, first_field=None, exclude_fields=(), n_significant_figures=4):
    """Write a CSV/TSV file.

    Parameters
    ----------

    data : list of dicts
    first_field : str
        The name of the field that should come first in the file.
    exclude_fields : list-like
        Fields present in the data that should not be saved in the file.
    n_significant_figures : int
        Number of significant figures used for floating-point numbers in the file.

    """
    path = Path(path)
    ensure_dir_exists(path.parent)
    delimiter = '\t' if path.suffix == '.tsv' else ','
    with path.open('w', newline='') as f:
        if not data:
            logger.info("Data was empty when writing %s.", path)
            return
        # Get the union of all keys from all rows.
        fields = set().union(*data)
        # Remove ignored fields.
        for field in exclude_fields:
            if field in fields:
                fields.remove(field)
        # Make sure the first field is the first one.
        if first_field in fields:
            fields.remove(first_field)
            fields = [first_field] + sorted(fields)
        else:
            fields = sorted(fields)
        writer = csv.writer(f, delimiter=delimiter)
        # Write the header.
        writer.writerow(fields)
        # Write all rows.
        writer.writerows(
            [[_pretty_floats(row.get(field, None), n_significant_figures)
             for field in fields] for row in data])
    logger.debug("Wrote %s.", path)


def _read_tsv_simple(path):
    """Read a CSV/TSV file with only two columns: cluster_id and <field>.

    Return (field_name, dictionary {cluster_id: value}).

    """
    path = Path(path)
    data = {}
    if not path.exists():
        logger.debug("%s does not exist, skipping.", path)
        return data
    # Find whether the delimiter is tab or comma.
    with path.open('r') as f:
        delimiter = '\t' if '\t' in f.readline() else ','
    with path.open('r') as f:
        reader = csv.reader(f, delimiter=delimiter)
        # Skip the header.
        _, field_name = next(reader)
        for row in reader:
            cluster_id, value = row
            cluster_id = int(cluster_id)
            data[cluster_id] = _try_make_number(value)
    logger.debug("Read %s.", path)
    return field_name, data


def _write_tsv_simple(path, field_name, data):
    """Write a CSV/TSV file with two columns: cluster_id and <field>.

    data is a dictionary {cluster_id: value}.

    """
    path = Path(path)
    ensure_dir_exists(path.parent)
    delimiter = '\t' if path.suffix == '.tsv' else ','
    with path.open('w', newline='') as f:
        writer = csv.writer(f, delimiter=delimiter)
        writer.writerow(['cluster_id', field_name])
        writer.writerows([(cluster_id, data[cluster_id]) for cluster_id in sorted(data)])
    logger.debug("Wrote %s.", path)


#------------------------------------------------------------------------------
# Various Python utility functions
#------------------------------------------------------------------------------

def _fullname(o):
    """Return the fully-qualified name of a function."""
    return o.__module__ + "." + o.__name__ if o.__module__ else o.__name__


def _load_from_fullname(name):
    """Load a Python object from its fully qualified name."""
    if not isinstance(name, str):
        return name
    parts = name.rsplit('.', 1)
    return getattr(import_module(parts[0]), parts[1], parts[1])


def _git_version():
    """Return the git version."""
    curdir = os.getcwd()
    os.chdir(str(Path(__file__).parent))
    try:
        with open(os.devnull, 'w') as fnull:
            version = ('-git-' + subprocess.check_output(
                       ['git', 'describe', '--abbrev=8', '--dirty', '--always', '--tags'],
                       stderr=fnull).strip().decode('ascii'))
            return version
    except (OSError, subprocess.CalledProcessError):  # pragma: no cover
        return ""
    finally:
        os.chdir(curdir)


def phy_config_dir():
    """Return the absolute path to the phy user directory. By default, `~/.phy/`."""
    return Path.home() / '.phy'


def ensure_dir_exists(path):
    """Ensure a directory exists, and create it otherwise."""
    path = Path(path)
    if path.exists():
        assert path.is_dir()
    else:
        path.mkdir(exist_ok=True, parents=True)
    assert path.exists() and path.is_dir()



# -*- coding: utf-8 -*-

"""ALF dataset generation."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging
from pathlib import Path
import shutil
import ast
import uuid

from tqdm import tqdm
import numpy as np

from phylib.utils._misc import _read_tsv_simple, ensure_dir_exists
from phylib.io.array import _spikes_per_cluster, _unique
from phylib.io.model import load_model

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# File utils
#------------------------------------------------------------------------------

NSAMPLE_WAVEFORMS = 500  # number of waveforrms sampled out of the raw data

_FILE_RENAMES = [  # file_in, file_out, squeeze (bool to squeeze vector from matlab in npy)
    ('params.py', 'params.py', None),
    ('cluster_KSLabel.tsv', 'cluster_KSLabel.tsv', None),
    ('spike_clusters.npy', 'spikes.clusters.npy', True),
    ('spike_templates.npy', 'spikes.templates.npy', True),
    ('channel_positions.npy', 'channels.localCoordinates.npy', False),
    ('channel_probe.npy', 'channels.probes.npy', True),
    ('channel_labels.npy', 'channels.labels.npy', True),
    ('cluster_probes.npy', 'clusters.probes.npy', True),
    ('cluster_shanks.npy', 'clusters.shanks.npy', True),
    ('whitening_mat.npy', '_kilosort_whitening.matrix.npy', False),
    ('_phy_spikes_subset.channels.npy', '_phy_spikes_subset.channels.npy', False),
    ('_phy_spikes_subset.spikes.npy', '_phy_spikes_subset.spikes.npy', False),
    ('_phy_spikes_subset.waveforms.npy', '_phy_spikes_subset.waveforms.npy', False),
    ('drift_depths.um.npy', 'drift_depths.um.npy', False),
    ('drift.times.npy', 'drift.times.npy', False),
    ('drift.um.npy', 'drift.um.npy', False),
    # ('cluster_group.tsv', 'ks2/clusters.phyAnnotation.tsv', False), # todo check indexing, add2QC
]

FILE_DELETES = [
    'temp_wh.dat',  # potentially large file that will clog the servers
]


def _read_npy_header(filename):
    d = {}
    with open(filename, 'rb') as fid:
        d['magic_string'] = fid.read(6)
        d['version'] = fid.read(2)
        d['len'] = int.from_bytes(fid.read(2), byteorder='little')
        d = {**d, **ast.literal_eval(fid.read(d['len']).decode())}
    return d


def _create_if_possible(path, new_path, force=False):
    """Prepare the copy/move/symlink of a file, by making sure the source exists
    while the destination does not."""
    if not Path(path).exists():  # pragma: no cover
        logger.warning("Path %s does not exist, skipping.", path)
        return False
    if Path(new_path).exists() and not force:  # pragma: no cover
        logger.warning("Path %s already exists, skipping.", new_path)
        return False
    ensure_dir_exists(new_path.parent)
    return True


def _copy_if_possible(path, new_path, force=False):
    if not _create_if_possible(path, new_path, force=force):
        return False
    logger.debug("Copying %s to %s.", path, new_path)
    shutil.copy(path, new_path)
    return True


def _load(path):
    path = str(path)
    if path.endswith('.npy'):
        return np.load(path)
    elif path.endswith(('.csv', '.tsv')):
        return _read_tsv_simple(path)[1]  # the function returns a tuple (field, data)
    elif path.endswith('.bin'):
        # TODO: configurable dtype
        return np.fromfile(path, np.int16)


#------------------------------------------------------------------------------
# Ephys ALF creator
#------------------------------------------------------------------------------

class EphysAlfCreator(object):
    """Class for converting a dataset in KS/phy format into ALF."""

    def __init__(self, model):
        self.model = model
        self.dir_path = Path(model.dir_path)
        self.spc = _spikes_per_cluster(model.spike_clusters)
        self.cluster_ids = _unique(self.model.spike_clusters)

    def convert(self, out_path, force=False, label='', ampfactor=1):
        """Convert from KS/phy format to ALF."""
        logger.info("Converting dataset to ALF.")
        self.out_path = Path(out_path)
        self.label = label
        self.ampfactor = ampfactor
        if self.out_path.resolve() == self.dir_path.resolve():
            raise IOError("The source and target directories cannot be the same.")
        if not self.out_path.exists():
            self.out_path.mkdir()

        with tqdm(desc="Converting to ALF", total=135) as bar:
            bar.update(10)
            self.make_cluster_objects()
            bar.update(10)
            self.make_channel_objects()
            bar.update(5)
            self.make_template_and_spikes_objects()
            bar.update(30)
            self.model.save_spikes_subset_waveforms(
                NSAMPLE_WAVEFORMS, sample2unit=self.ampfactor)
            bar.update(50)
            self.make_depths()
            bar.update(20)
            self.rm_files()
            bar.update(10)
            self.copy_files(force=force)
            bar.update(10)
            self.rename_with_label()
            bar.update(10)
            self.compress_spikes_dtypes()

        # Return the TemplateModel of the converted ALF dataset if the params.py file exists.
        params_path = self.out_path / 'params.py'
        if params_path.exists():
            return load_model(params_path)

    def copy_files(self, force=False):
        for fn0, fn1, squeeze in _FILE_RENAMES:
            f0 = self.dir_path / fn0
            f1 = self.out_path / fn1
            _copy_if_possible(f0, f1, force=force)
            if f0.exists() and squeeze and f0.suffix == '.npy':
                h = _read_npy_header(f0)
                # ks2 outputs vectors as multidimensional arrays. If there is no distinction
                # for Matlab, there is one in Numpy
                if len(h['shape']) == 2 and h['shape'][-1] == 1:  # pragma: no cover
                    d = np.load(f0)
                    np.save(f1, d.squeeze())
                    continue

    def rm_files(self):
        for fn0 in FILE_DELETES:
            fn = self.dir_path.joinpath(fn0)
            if fn.exists():  # pragma: no cover
                fn.unlink()

    # File creation
    # -------------------------------------------------------------------------

    def _save_npy(self, filename, arr, dtype=None):
        """Save an array into a .npy file."""
        dtype = arr.dtype if dtype is None else dtype
        np.save(self.out_path / filename, arr.astype(dtype))

    def make_cluster_objects(self):
        """Create clusters.channels, clusters.waveformsDuration and clusters.amps"""
        peak_channel_path = self.dir_path / 'clusters.channels.npy'
        if not peak_channel_path.exists():
            # self._save_npy(peak_channel_path.name, self.model.templates_channels)
            self._save_npy(peak_channel_path.name, self.model.clusters_channels)

        waveform_duration_path = self.dir_path / 'clusters.peakToTrough.npy'
        if not waveform_duration_path.exists():
            # self._save_npy(waveform_duration_path.name, self.model.templates_waveforms_durations)
            waveform_duration = self.model.clusters_waveforms_durations
            waveform_duration[self.model.nan_idx] = np.nan
            self._save_npy(waveform_duration_path.name, waveform_duration)

        # group by average over cluster number
        # camps = np.zeros(self.model.templates_channels.shape[0],) * np.nan
        camps = np.zeros(self.model.clusters_channels.shape[0], ) * np.nan
        camps[self.cluster_ids] = self.model.clusters_amplitudes
        amps_path = self.dir_path / 'clusters.amps.npy'
        self._save_npy(amps_path.name, camps * self.ampfactor)

        # clusters uuids
        uuid_list = ['uuids']
        uuid_list.extend([str(uuid.uuid4()) for _ in range(camps.size)])
        with open(self.out_path / 'clusters.uuids.csv', 'w+') as fid:
            fid.write('\n'.join(uuid_list))

    def make_channel_objects(self):
        """If there is no rawInd file, create it"""
        rawInd_path = self.dir_path / 'channels.rawInd.npy'
        rawInd = np.zeros_like(self.model.channel_probes).astype(int)
        channel_offset = 0
        for probe in np.unique(self.model.channel_probes):
            ind = self.model.channel_probes == probe
            rawInd[ind] = self.model.channel_mapping[ind] - channel_offset
            channel_offset += np.max(self.model.channel_mapping[ind])
        self._save_npy(rawInd_path.name, rawInd)

    def make_depths(self):
        """Make spikes.depths.npy, clusters.depths.npy."""
        channel_positions = self.model.channel_positions
        assert channel_positions.ndim == 2

        spike_clusters = self.model.spike_clusters
        assert spike_clusters.ndim == 1

        cluster_channels = np.load(self.out_path / 'clusters.channels.npy')
        assert cluster_channels.ndim == 1
        n_clusters = cluster_channels.shape[0]

        clusters_depths = channel_positions[cluster_channels, 1]
        clusters_depths[self.model.nan_idx] = np.nan
        assert clusters_depths.shape == (n_clusters,)

        if self.model.sparse_features is None:
            spikes_depths = clusters_depths[spike_clusters]
        else:
            spikes_depths = self.model.get_depths()
        self._save_npy('spikes.depths.npy', spikes_depths, np.float32)
        self._save_npy('clusters.depths.npy', clusters_depths)

    def make_template_and_spikes_objects(self):
        """Creates the template waveforms sparse object
        Without manual curation, it also corresponds to clusters waveforms objects.
        """
        # "We cannot just rename/copy spike_times.npy because it is in unit of samples,
        # and not seconds
        self._save_npy('spikes.times.npy', self.model.spike_times)
        self._save_npy('spikes.samples.npy', self.model.spike_samples)
        spike_amps, templates_v, template_amps = self.model.get_amplitudes_true(self.ampfactor,
                                                                                use='templates')
        self._save_npy('spikes.amps.npy', spike_amps, np.float32)
        self._save_npy('templates.amps.npy', template_amps)

        if self.model.sparse_templates.cols:
            raise NotImplementedError("Sparse template export to ALF not implemented yet")
        else:
            n_templates, n_wavsamps, nchall = templates_v.shape
            # for some datasets, 32 may be too much
            ncw = min(self.model.n_closest_channels, nchall)
            assert n_templates == self.model.n_templates
            templates = np.zeros((n_templates, n_wavsamps, ncw), dtype=np.float32)
            templates_inds = np.zeros((n_templates, ncw), dtype=np.int32)
            # for each template, find the nearest channels to keep (one the same probe...)
            for t in np.arange(n_templates):
                current_probe = self.model.channel_probes[self.model.templates_channels[t]]
                channel_distance = np.sum(np.abs(
                    self.model.channel_positions -
                    self.model.channel_positions[self.model.templates_channels[t]]), axis=1)
                channel_distance[self.model.channel_probes != current_probe] += np.inf
                templates_inds[t, :] = np.argsort(channel_distance)[:ncw]
                templates[t, ...] = templates_v[t, :][:, templates_inds[t, :]]
            np.save(self.out_path.joinpath('templates.waveforms'), templates)
            np.save(self.out_path.joinpath('templates.waveformsChannels'), templates_inds)

            _, clusters_v, cluster_amps = self.model.get_amplitudes_true(self.ampfactor,
                                                                         use='clusters')
            n_clusters, n_wavsamps, nchall = clusters_v.shape
            # for some datasets, 32 may be too much
            ncw = min(self.model.n_closest_channels, nchall)
            assert n_clusters == self.model.n_clusters
            templates = np.zeros((n_clusters, n_wavsamps, ncw), dtype=np.float32)
            templates_inds = np.zeros((n_clusters, ncw), dtype=np.int32)
            # for each template, find the nearest channels to keep (one the same probe...)
            for t in np.arange(n_clusters):
                channels = self.model.clusters_channels

                current_probe = self.model.channel_probes[channels[t]]
                channel_distance = np.sum(np.abs(
                    self.model.channel_positions -
                    self.model.channel_positions[channels[t]]), axis=1)
                channel_distance[self.model.channel_probes != current_probe] += np.inf
                templates_inds[t, :] = np.argsort(channel_distance)[:ncw]
                templates[t, ...] = clusters_v[t, :][:, templates_inds[t, :]]
            np.save(self.out_path.joinpath('clusters.waveforms'), templates)
            np.save(self.out_path.joinpath('clusters.waveformsChannels'), templates_inds)
            np.save(self.out_path.joinpath('clusters.amps'), cluster_amps)

    def rename_with_label(self):
        """add the label as an ALF part name before the extension if any label provided"""
        if not self.label:
            return
        glob_patterns = ['channels.*', 'clusters.*', 'spikes.*', 'templates.*']
        for pattern in glob_patterns:
            for f in self.out_path.glob(pattern):
                f.rename(f.with_suffix(f'.{self.label}{f.suffix}'))

    def compress_spikes_dtypes(self):
        """Convert clusters and templates to int16."""
        for attribute in ['templates', 'clusters']:
            fn = next(self.out_path.glob(f'spikes.{attribute}.*npy'))
            np.save(fn, np.load(fn).astype(np.uint16))
